/*
  Copyright (c) 2014-2015, Matthias Schiffer <mschiffer@universe-factory.net>
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.
    2. Redistributions in binary form must reproduce the above copyright notice,
       this list of conditions and the following disclaimer in the documentation
       and/or other materials provided with the distribution.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/


#pragma once


#include "TimeProvider.hpp"

#include <chrono>
#include <condition_variable>
#include <cstdint>
#include <deque>
#include <functional>
#include <limits>
#include <queue>
#include <thread>


namespace RPGEdit {

namespace Control {

class EventBus {
public:
	typedef std::function<void ()> Event;
	typedef std::pair<uint64_t, Event> EventEntry;

private:
	static bool compare_events(const EventEntry &e1, const EventEntry &e2) {
		return e1.first > e2.first;
	}

	std::priority_queue<EventEntry, std::deque<EventEntry>, bool (*)(const EventEntry &, const EventEntry &)> events;

	std::mutex mutex;
	std::condition_variable cond;

public:
	EventBus() : events(compare_events) {
	}

	void enqueue(const Event &event, uint64_t time) {
		std::lock_guard<std::mutex> lock(mutex);
		events.push(EventEntry(time, event));
		cond.notify_one();
	}

	EventEntry get(TimeProvider *timeProvider) {
		std::unique_lock<std::mutex> lock(mutex);

		while (true) {
			if (events.empty()) {
				cond.wait(lock);
				continue;
			}

			EventEntry top = events.top();
			uint64_t time = timeProvider->now();

			if (top.first > time) {
				cond.wait_for(lock, std::chrono::milliseconds(top.first - time));
				continue;
			}

			events.pop();
			return top;
		}
	}

	uint64_t peek() {
		std::lock_guard<std::mutex> lock(mutex);

		if (events.empty())
			return std::numeric_limits<uint64_t>::max();

		return events.top().first;
	}
};

}

}
