/*
  Copyright (c) 2014-2015, Matthias Schiffer <mschiffer@universe-factory.net>
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.
    2. Redistributions in binary form must reproduce the above copyright notice,
       this list of conditions and the following disclaimer in the documentation
       and/or other materials provided with the distribution.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/


#include "RPGEdit.hpp"
#include "MapContext.hpp"
#include "../view/MapView.hpp"


namespace RPGEdit {

namespace Control {

bool RPGEdit::handleSystemEvent(const SDL_Event &event) {
	uint64_t time = timeProvider.now();

	switch (event.type) {
	case SDL_KEYDOWN:
		eventBus.enqueue([=] { inputHandler.keyPressed(event.key.keysym.scancode, time); }, time);
		break;

	case SDL_KEYUP:
		eventBus.enqueue([=] { inputHandler.keyReleased(event.key.keysym.scancode, time); }, time);
		break;

	case SDL_QUIT:
		return false;
	}

	return true;
}

void RPGEdit::systemLoop() {
	const int MIN_FRAME_DELAY = 10;

	uint32_t lastFrameTicks = SDL_GetTicks();

	while (true) {
		uint32_t diff = SDL_GetTicks() - lastFrameTicks;
		int timeout = std::max(MIN_FRAME_DELAY - int(diff), 0);

		SDL_Event event;
		if (SDL_WaitEventTimeout(&event, timeout)) {
			if (!handleSystemEvent(event))
				return;

			continue;
		}

		lastFrameTicks = SDL_GetTicks();

		SDL_RenderClear(window->getRenderer());

		{
			std::unique_lock<std::mutex> lock(modelMutex);

			ctx->render(std::min(timeProvider.now(), handledTime));
		}

		SDL_RenderPresent(window->getRenderer());
	}
}

void RPGEdit::eventLoop() {
	while (true) {
		EventBus::EventEntry event = eventBus.get(&timeProvider);

		if (!event.second)
			return;

		{
			std::lock_guard<std::mutex> lock(modelMutex);

			event.second();

			handledTime = eventBus.peek();
		}
	}
}

void RPGEdit::run() {
	std::unique_ptr<Model::Map> map = Model::Map::load("test");

	window = std::make_shared<View::Window>();

	ctx = std::make_shared<MapContext>(&eventBus, &inputHandler, &scriptContext, window, *map);

	eventThread = std::thread([this] { eventLoop(); });

	systemLoop();

	eventBus.enqueue(EventBus::Event(), timeProvider.now());
	eventThread.join();
}

}

}
