/*
  Copyright (c) 2014-2015, Matthias Schiffer <mschiffer@universe-factory.net>
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.
    2. Redistributions in binary form must reproduce the above copyright notice,
       this list of conditions and the following disclaimer in the documentation
       and/or other materials provided with the distribution.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/


#pragma once

extern "C" {
#include <lua.h>
#include <lualib.h>
#include <lauxlib.h>
}

#include <memory>
#include <string>
#include <unordered_map>

namespace RPGEdit {

namespace Model {

class ScriptValue {
public:
	virtual void push(lua_State *L) = 0;

	virtual ~ScriptValue() {}
};


class ScriptBoolean : public ScriptValue {
private:
	bool value;

public:
	ScriptBoolean(bool value0) : value(value0) {
	}

	ScriptBoolean & operator=(bool newValue) {
		value = newValue;
		return *this;
	}

	operator bool() const {
		return value;
	}

	virtual void push(lua_State *L) {
		lua_pushboolean(L, value);
	}
};

class ScriptNumber : public ScriptValue {
private:
	lua_Number value;

public:
	ScriptNumber(lua_Number value0) : value(value0) {
	}

	ScriptNumber & operator=(lua_Number newValue) {
		value = newValue;
		return *this;
	}

	operator lua_Number() const {
		return value;
	}

	virtual void push(lua_State *L) {
		lua_pushnumber(L, value);
	}
};

class ScriptString : public ScriptValue {
private:
	std::string value;

public:
	ScriptString(std::string &value0) : value(value0) {
	}

	ScriptString & operator=(std::string &newValue) {
		value = newValue;
		return *this;
	}

	operator std::string &() {
		return value;
	}

	virtual void push(lua_State *L) {
		lua_pushstring(L, value.c_str());
	}
};

class ScriptTable : public ScriptValue {
public:
	typedef std::unordered_map<std::shared_ptr<ScriptValue>, std::shared_ptr<ScriptValue>> MapType;

private:
	MapType value;

public:
	std::shared_ptr<ScriptValue> & operator[](const std::shared_ptr<ScriptValue> &key) {
		return value[key];
	}

	virtual void push(lua_State *L) {
		lua_createtable(L, 0, value.size());

		for (const auto &entry : value) {
			entry.first->push(L);
			entry.second->push(L);

			lua_settable(L, -3);
		}
	}
};

}

}
